// Vector.h: interface and implementation for the Vector class template.
//
// Friday Nov 19th 2004
//
// CSCIE 225  Russell Lowke
//
//////////////////////////////////////////////////

#if !defined(VECTOR_H_INCLUDED)
#define VECTOR_H_INCLUDED

#include <iostream>
using namespace std;

// Forward declare class template and function template
template <class T> class Vector;
template <class T> ostream& operator<<(ostream& os, Vector<T>& v);

template <class T>
class Vector{
	
	friend ostream& operator<< <>(ostream& os, Vector<T>& v);
	
	public:
		Vector<T>();								// default constructor
		~Vector<T>();								// destructor
		void   clear();								// reset the m_stack
		size_t size();								// get the m_stack size
		void   push(T n);							// push int onto m_stack
		void   push(T a[], size_t array_size);		// push array onto m_stack
		size_t memUsage() const;
		T	   sum() const;							// sum all values
		T      operator[](size_t index) const;		// return item at given index
		void   setAt(size_t index,					// stores item at given index
				     const T& item);
		void   setAtGrow(size_t index,				// stores item at given index and resizes if ness.
						 const T& item);
		void   minimize();							// reduces the memory used by object.

	private:
	
		void   doubleCapacity();					// double the capacity of the m_stack

		size_t m_max_size;							// the capacity of m_stack
		size_t m_top_of_stack;						// next free item on m_stack
		T*   m_stack;								// points to the m_stack of Ts
		
};

////////////
//
//  Vector 
//


//
// default constructor
//
template <class T>
Vector<T>::Vector() {
	
	m_max_size = 0;
	m_top_of_stack = 0;
	m_stack = new T[m_max_size];
}

//
// destructor
//
template <class T>
Vector<T>::~Vector() {
	delete []m_stack;
}

//
// reset the m_stack
//
template <class T>
void Vector<T>::clear() {

	// erase all data in m_stack
	delete []m_stack;
	
	// reset the Vector
	m_max_size = 0;
	m_top_of_stack = 0;
	m_stack = new T[m_max_size];
}

//
// push value n onto the m_stack
//
template <class T>
void Vector<T>::push(T n) {

	// if full, double the capacity
	if (m_top_of_stack >= m_max_size) {
		doubleCapacity();
	}
	
	// put n on m_stack
	m_stack[m_top_of_stack] = n;
		
	// increment m_top_of_stack counter
	m_top_of_stack += 1;
}

//
//  push array a onto the m_stack
//
template <class T>
void Vector<T>::push(T a[], size_t n) {

	// iterate through a[], pushing values
	for (size_t i = 0; i < n; ++i) {
		push(a[i]);
	}
}

//
//  doubles the capacity of IntStack.
//
template <class T>
void Vector<T>::doubleCapacity() {

	size_t new_size = m_max_size * 2;					// double capacity

	if (m_max_size == 0) {								// special case for 0
		new_size = 1;
	}

	T* dbl_stack = new T[new_size];						// make a double size m_stack
	for(size_t i = 0; i < m_max_size; ++i) {			// copy m_stack values into it
		dbl_stack[i] = m_stack[i];
	}

	delete []m_stack;									// release the old m_stack, then
	m_stack = dbl_stack;								// reassign to dbl_stack
	m_max_size = new_size;								// set the new m_max_size
}

//
// return the stack size
//
template <class T>
size_t Vector<T>::size() {
	return m_top_of_stack;
}

//
// return the memory size of the Vector
//
template <class T>
size_t Vector<T>::memUsage() const {

	if (m_max_size == 0) {
		return 0;
	} else {
		return ( sizeof(m_stack[0]) * m_max_size );
	}

}

template <class T>
T Vector<T>::sum() const {
	T sum = T();

	// iterate through all items on m_stack
	for (size_t i = 0; i < m_top_of_stack ; ++i) {
		sum += m_stack[i];
	}

	return sum;
}

//
// operator[] returns item at given index
//
template <class T>
T Vector<T>::operator[](size_t index) const {

	// error checking
	if ( (index < 0) ||
		 (index >= m_top_of_stack) ) {
		cerr << "[" << (unsigned int) index << "] out of range" << endl;
		return T();
	}

	return m_stack[index];
}

//
// stores item at given index
//
template <class T>
void Vector<T>::setAt(size_t index, const T& item) {

	// error checking
	if ( (index < 0) ||
		 (index >= m_max_size) ) {
		cerr << (unsigned int) index << " out of range" << endl;
		return;
	}

	// expand up to index
	if ( index >= m_top_of_stack ) {
		for (size_t i = m_top_of_stack; i < index ; ++i) {
			m_stack[i] = T();
		}
		m_top_of_stack = index + 1;
	}

	m_stack[index] = item;
}

//
// stores item at given index and resizes if ness.
//
template <class T>
void Vector<T>::setAtGrow(size_t index, const T& item) {

	// ensure index within bounds
	while (index >= m_max_size) {
		doubleCapacity();
	}

	setAt(index, item);
}

//
// reduces the memory used by object.
//
template <class T>
void Vector<T>::minimize() {
	
	size_t stack_size = m_top_of_stack - 1;				// last item is m_top_of_stack - 1

	if (m_top_of_stack == 0) {							// correct for empty stack
		stack_size = 0;
	}

	T* min_stack = new T[stack_size];					// make a minim size stack
	for(size_t i = 0; i < stack_size; ++i) {			// copy m_stack values into it
		min_stack[i] = m_stack[i];
	}

	delete []m_stack;									// release the m_stack, then
	m_stack    = min_stack;								// reassign to min_stack
	m_max_size = stack_size;							// set the new m_max_size

	// NOTE: this can leave m_max_size one smaller than m_top_of_stack.
	//       as m_top_of_stack points to location after last item.

}

//
//	print the Vector
//
template <class T>
ostream& operator<< (ostream& os, Vector<T>& v) {

	// find the number of items in m_stack
	size_t n = v.m_top_of_stack;
	
	os << "size: " << (unsigned int) v.size() << endl;
	os << " max: " << (unsigned int) v.m_max_size << endl;
	// iterate through all items on m_stack
	for (size_t i = 0; i < n ; ++i) {
		os << "[" << (unsigned int) i << "] " << v.m_stack[i] << endl;
	}

	os << endl;

	return os;
}

#endif // !defined(VECTOR_H_INCLUDED)