// account.cpp: implementation of the Account class.
//
//////////////////////////////////////////////////////////////////////

#include "account.h"
#include <iostream>

using std::cout;
using std::endl;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Account_base::Account_base() {}
Account_base::~Account_base() {}
Account::Account( const Market& m ) : Account_base(), m_market( m ), m_cashBalance( 0 ) {}
Account::~Account() {}

std::ostream& operator<< ( std::ostream& os, const Account& acct ) {

	os << "Account with cash balance of $" << acct.m_cashBalance;
	Account::HoldingIter hi = acct.m_holdings.begin();
	if ( hi == acct.m_holdings.end() ) {
		os << ", holding no stock." << endl;
	} else {
		os << endl << "  Holdings" << endl;
		for ( ; hi != acct.m_holdings.end(); hi++ ) {
			os << hi->second << " shares of " << hi->first << endl;
		}
	}

	return os;
}


bool Account::sell( const Symbol& s, unsigned int nShares ) {
	unsigned int price;
	if ( ! m_market.getPrice( s, &price ) )
		throw UnknownSymbol(s);

	// Avoid using [] here to avoid inserting a 0 holding
	HoldingIter hi = m_holdings.find(s);
	if ( hi == m_holdings.end() || hi->second < nShares )
		InsuffShares( s, hi->second );

	unsigned int yield = nShares * price;
	m_holdings[s] -= nShares;
	if ( m_holdings[s] == 0 ) {
		m_holdings.erase( s );
	}

	m_cashBalance += yield;
	return true; 
}


bool Account::buy( const Symbol& s, unsigned int nShares ) {

	unsigned int price;
	if ( ! m_market.getPrice( s, &price ) )
		throw UnknownSymbol(s);

	unsigned int cost = nShares * price;
	if ( cost > m_cashBalance )
		throw InsuffFunds( m_cashBalance, cost );

	m_cashBalance -= cost;
	m_holdings[s] += nShares;

	return true; 
}


bool Account::deposit( unsigned int amount ) {
	m_cashBalance += amount;
	return true; 
}


bool Account::withdraw( unsigned int amount ) {
	if ( amount > m_cashBalance )
		throw InsuffFunds( m_cashBalance, amount );

	m_cashBalance -= amount;
	return true;
}

void Account::processInterest() {} // no interest

GoldAccount::GoldAccount( const Market& m ) : Account( m ) {}
GoldAccount::~GoldAccount() {}

void GoldAccount::processInterest() {

	// award 0.1% interest
	double interest = m_cashBalance * 0.001;
	// Note: typecast here will lose cents
	m_cashBalance += (unsigned int) interest;
}

// constructor for platinum calls constructor Gold calls constructor Account
PlatinumAccount::PlatinumAccount( const Market& m ) : GoldAccount( m ), m_credit_line(0) {}

PlatinumAccount::~PlatinumAccount() {}

bool PlatinumAccount::payBack(unsigned int amount) {
	// pay off the credit balance from the cash balance.

	if (amount > m_credit_line)
		amount = m_credit_line;			// too much, they don't owe that much!

	if (amount > m_cashBalance)			// can't pay it, they don't have that much!
		throw InsuffFunds( m_cashBalance, amount );

	m_cashBalance -= amount;			// remove cash amount from account
	m_credit_line -= amount;			// remove cash amount from debt

	return true;
}

void PlatinumAccount::processInterest() {

	// add interest the same way as the Gold account
	GoldAccount::processInterest();

	// charge 1.8% on the credit line
	double creditCharge = m_credit_line * 0.018;
	m_credit_line += (unsigned int) creditCharge;	// Note: typecast here will lose some cents.
}

// borrow from credit to buy shares if nessesary
bool PlatinumAccount::buy( const Symbol& s, unsigned int nShares ) {

	unsigned int price;
	if ( ! m_market.getPrice( s, &price ) )
		throw UnknownSymbol(s);

	unsigned int cost = nShares * price;
	if ( cost > m_cashBalance ) {
		unsigned int amountNeeded = cost - m_cashBalance;
		m_credit_line += amountNeeded;		// add cash borrowed to credit line
		m_cashBalance += amountNeeded;		// add cash borrowed to account
	}

	// buy the shares
	return Account::buy(s, nShares );
}