
//////////////////////////////////////////////////////////////////////
//
// String.cpp: implementation of the String class.
//
//////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <string.h>
#include "String.h"

using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

String::String() {
	init();								// default constructor called init
}

String::String(const char* s) {
	init();								// supplied constructor that creates
	size_t len = strlen(s);				// String from a char pointer
	ensureSize( len );
	memcpy( m_buffer, s, len );
	m_length = len;
}

String::String(const String& s) {
	*this = s;							// work of copy constructor
										// done by operator =
}

String::String(char ch, unsigned int count /* = 1 */ ) {
	init();								// supplied constructor that creates
	ensureSize( count );				// a string of length count with
	memset( m_buffer, ch, count );		// all values initialized to ch
	m_length = count;
}

String::~String() {
	delete[] m_buffer;					// supplied destructor
}

size_t String::length() const {
	return m_length;					// public get m_length
}


//
// append to this String the String other
//
String& String::append(const String& other) {

	// calculate resultant string length 
	size_t o_length = other.length();
	size_t n_length = m_length + o_length;

	// ensure buffersize can cope
	ensureSize(n_length);

	// copy chars from other into this
	for (size_t i = 0; i < o_length; ++i) {
		m_buffer[m_length + i] = other.m_buffer[i];
	}

	// update string size
	m_length = n_length;

	return *this;		// return this String
}

//
// special case append when other is a char*
//
String& String::append(const char* other) {
	String temp_str(other);

	// now append temp_str
	return append(temp_str);
}

//
// create a new substring starting at start and numChars long
//
String String::substr( size_t start, size_t numChars) const {
   // NOTE FROM ERIC: size_t is an unsigned type, so it can never
   // be less than 0.
	// error checking
	if ( (start < 0) ||						// is start within string?
		 (start > m_length) ) {
		cerr << "'start' out of range with value of " << (unsigned int) start << 
			    " passed to substr()" << endl;
		start = numChars = 0;				// make start legal.

	} else if (numChars < 0) {				// is numChars is legal?
      // NOTE FROM ERIC: unreachable code for the same reason
		numChars = 0;						// make numChars 0.
	} else if ( (numChars > m_length) ||		// prevent busting max int...
		        (numChars + start > m_length) ) {
		numChars = m_length - start;		// make numChars legal.
	}

	// initialize a new string of length numChars
	String n_str('0', numChars);

	// copy chars from this into n_str
	for (size_t i = 0; i < numChars; ++i) {
		n_str.m_buffer[i] = m_buffer[start + i];
	}

	// assign the new strings length
	n_str.m_length = numChars;

	return n_str;
}

//
// return character in this at position pos
//
char String::char_at( size_t pos ) const {

	// error checking
	if ( (pos < 0) ||
		 (pos >= m_length) ) {
		cerr << "'pos' out of range with value of " << (unsigned int) pos << 
			    " passed to char_at()" << endl;
		return '\0';
	}

	return m_buffer[pos];
}

//
// operator[] that performs char_at()
//
char String::operator[](size_t pos) const {
	return char_at(pos);
}

//
// operator+= that appends RHS to LHS using append()
//
String& String::operator+=(const String& other) {
	return append(other);
}

//
// operator= called by constructor.  Creates String initializing with other
//
String& String::operator=(const String& other) {
	if (this == &other) {
		return *this;									// allows for str = str;
	}

	init();												// initialize String
	size_t len = other.m_length;						// get the size of other
	ensureSize( len );									// create buffer
	memcpy( m_buffer, other.m_buffer, len );			// copy the memory from other
	m_length = len;										// assign new length

	return *this;
}

//
//	operator==  comparison that confirms if two Strings are equal
//
bool String::operator==(const String& other) const {

	if (other.m_length != m_length) {					// strings same length?
		return false;
	}
	
	for (size_t i = 0; i < m_length; ++i) {				// loop through each char
		if (m_buffer[i] != other.m_buffer[i]) {
			return false;
		}
	}
	return true;										// ok, strings are same
}

//
// operator<  less than test for case-sensitive ordering of Strings
//
bool String::operator< (const String& other) const {

	size_t o_length = other.length();					// find length of other

	// loop with i < length of other
	for (size_t i = 0; i < o_length; ++i) {

		if (i >= m_length) {
			return true;								// is shorter than other
		}

		// compare values
		if (m_buffer[i] < other.m_buffer[i]) {			// is smaller
			return true;
		} else if (m_buffer[i] > other.m_buffer[i]) {	// is larger
			return false;
		} else {										// is equal
			// values are equal, get next char
		}
	}

	return false;										// is equal or longer than other
}


//
//	supplied print String
//
std::ostream& operator<< (std::ostream& os, String s) {
	s.ensureSize( s.m_length + 1 );
	s.m_buffer[s.m_length] = 0;
	os << "S(" << (unsigned int) s.m_length << ","
	   << (unsigned int) s.m_bufSize << ")[" << s.m_buffer << "]";
	return os;
}


//
// supplied initializer
//
void String::init() {
	m_buffer = NULL;
	m_length = 0;
	m_bufSize = 0;
}

//
// supplied ensureSize() method allocating buffer space
//
void String::ensureSize( size_t size ) {
	if ( size <= m_bufSize ) {
		return;
	}

	// ensure buffer is of at least size
	char* pNew = new char[size];
	memcpy( pNew, m_buffer, m_length );
	m_bufSize = size;
	delete[] m_buffer;
	m_buffer = pNew;
}
